#ifndef __CHARACTRON_H__
#define __CHARACTRON_H__

#define CHARACTRON_FG 1
#define CHARACTRON_BG 0

typedef struct charactron
{
  SDL_Surface *font;    /* 256 x 256 grid of ASCII characters   */
  SDL_Surface *output;  /* surface that is drawn on             */

  Sint16 x,y;           /* position of x and y cursor           */
  Sint16 maxx,maxy;     /* width and height of output           */

  Uint32 fg,bg;         /* Cached fg/bg colors for filling      */
  Uint32 flag;
} charactron;

#define CH_ENABLE               0x80000000

#define CHARACTRON_FONT_W(c)    ((c)->font->w >> 4)
#define CHARACTRON_FONT_H(c)    ((c)->font->h >> 4)
#define CHARACTRON_W_FREE(c)    ((c)->maxx - (c)->x)
#define CHARACTRON_H_FREE(c)    ((c)->maxy - (c)->y)

#define CHARACTRON_ENABLE(c)	do { (c)->flag |= CH_ENABLE; } while(0)
#define CHARACTRON_DISABLE(c)	do { (c)->flag &= ~CH_ENABLE; } while(0)

#define CHARACTRON_ENABLED(c)   ((c)->flag & CH_ENABLE)

void charactron_free(charactron *c);
charactron *charactron_load(SDL_RWops *img, SDL_Surface *out, int autoclose);

int charactron_draw_letter(charactron *c, Uint8 letter);
int charactron_draw_data(charactron *c, const char *str, long len);
int charactron_linefeed(charactron *c);
int charactron_formfeed(charactron *c);

int charactron_set_color(charactron *c, int i, SDL_Color cl);

#endif/*__CHARACTRON_H__*/
